local A = select(2, ...);
local M = A.Lfg;
local D = A.Api.Dungeon;
local U = A.Utils;
local dungeons = {}
local fb = A.Lib.FB;
local C = A.Constants;
local P = A.Player;
local R = A.Request;
local L = A.Localize.translate;

local Bus = A.ChatBus;
local TypeApi = A.Api.Type;
local RoleApi = A.Api.Role;
local SEND_DELAY_IN_SECONDS = 20;
local Cache = A.Cache;

local ChatThrottler = U:CreateThrottler(function(message) return message.time end, storeForInSeconds, cleanEachInSeconds);


local LOC = {
    SelectDungeon = L("Select dungeon"),
    NotSpecified = L("Not Specified"),
    Empty = L("---"),
    Close = L("Close"),
    Role = L("Specify your role"),
    DefaultRole = L("Your role"),
    DefaultNote = L("Your note"),
    EditNote = L("Edit note"), 
    Change = L("Change"),
    Send = L("Send")
}


local MODECOLORS = {
    [true] = { 1, .5, 0 },
    [false] = { .6, 1, .5 }
}


local DungeonFilter = {
    _values = {},
    _changeListeners = {},

    Set = function(self, key, value)
        local old = self._values[key];
        self._values[key] = value;

        if old ~= value then
            self:TriggerChange();
        end
    end,

    Get = function(self, key)
        if key == nil then return self._values end
        return self._values[key];
    end,

    TriggerChange = function(self)

        if not self._triggerChange then
            self._triggerChange = U:Debounce(function() self:OnChange() end, .1)
        end
        self._triggerChange();
    end,

    OnChange = function(self)

        self.changed = true;
        for _i, cb in pairs(self._changeListeners) do

            cb(self._values);
        end
        self.changed = false;
    end,

    AddListener = function(self, cb)
        table.insert(self._changeListeners, cb);
    end,
}

local LFGDB = {
    items = {},
    itemsById = {},
    Init = function(self)
        if self._initialized then return end
        for _, dung in pairs(D:GetAll()) do
            local lfmdungeon = {};
            Mixin(lfmdungeon, dung);

            lfmdungeon.checked = Cache:GetOption("lfg.dungeon."..lfmdungeon.id..".checked");
            lfmdungeon.role = Cache:GetOption("lfg.dungeon."..lfmdungeon.id..".role");
            lfmdungeon.note = Cache:GetOption("lfg.dungeon."..lfmdungeon.id..".note");

            --print("# ", lfmdungeon.id, lfmdungeon.checked);

            table.insert(self.items, lfmdungeon);
            self.itemsById[lfmdungeon.id] = lfmdungeon;
        end
        self._initialized = true;
    end,

    GetCheckedHash = function(self)
        local res = {};
        for _i, item in pairs(self.items) do
            if item.checked then
                res[item.id] = true;
            end
        end
        return res;
    end,

    IsCheckedById = function(self, id)
        local dng = self:GetById(id);
        return dng.checked == true;
    end,

    Get = function(self, filters)

        self:Init();

        return D:FilterTable(self.items, filters);
    end,
    GetById = function(self, id)
        return self.itemsById[id];
    end
};

function M:GetCheckedDungeonsHash()
    return LFGDB:GetCheckedHash();
end

local FrameMixin = {

    SendRequest = function(self, data)

        local send = {
            type = C.Lfg,
            zone = data.name,
            mode = data.mode,
            role = data.role or P.role,
            note = data.note or P.note,
        }
        
        local text = R:ToString(send);
        Bus:SendChatMessage(text);

    end,

    GetData = function(self)
        if not self.data or DungeonFilter.changed then 
            DungeonFilter.changed = false;
            self.data = LFGDB:Get(DungeonFilter:Get())
        end
        return self.data;
    end,

    OnEvent = function(self, event, ...)
        
        if event == "CHAT_MSG_ADDON" then
            --print(event, ...);
        elseif event == "CHAT_MSG_CHANNEL" or event == "CHAT_MSG_SAY" then
            self:OnChatMessage(...);
        elseif event == "PLAYER_ENTERING_WORLD" then
            local isLogin, isReload = ...;
            self:OnPlayerEnteringWorld(isLogin, isReload)
        end
    end,

    OnPlayerEnteringWorld = function(self, isLogin, isReload)


        -- C_ChatInfo.SendAddonMessage(AddonPrefix, "You can't see this message", "WHISPER", playerName)
        -- C_ChatInfo.SendAddonMessage(AddonPrefix, "Hello world!", "WHISPER", playerName)
    end,

    OnChatMessage = function(self, text, playerName, languageName, channelName, playerName2, specialFlags, zoneChannelID, channelIndex, channelBaseName, unused, lineID, guid, bnSenderID, isMobile, isSubtitle, hideSenderInLetterbox, supressRaidIcons)
        --print(playerName, languageName, channelName, playerName2, specialFlags, zoneChannelID, channelIndex, channelBaseName, unused, lineID, guid, bnSenderID, isMobile, isSubtitle, hideSenderInLetterbox, supressRaidIcons);
        
        if P:IsMyGuid(guid) then return end;

        local request = U:TextSplit(text or "", " ");

        -- if request[1] == C.Lfg then

        --     -- someone is looking for group
        --     ProcessLookingForGroup(request);
        --     return;

        -- end

        if string.lower(request[1]) == string.lower(C.Lfm) then

            request = R:Parse(request);
            request.playerName = playerName;

            -- some group is looking for members
            self:ProcessLookingForMembers(request);

            return;

        end

    end,

    -- ProcessLookingForMembers = function(self, request)
    --     local dungeon = self:GetDungeonByRequest(request);
    --     if not dungeon then return end;
    --     if not self:IsMyRoleSuitable(request, dungeon) then return end
    --     local text = P:LfmResponse(dungeon);
    --     --Bus:SendAddonMessage(text, request)
    --     --SendChatMessage(text , "WHISPER" , nil , request.playerName);
    --     --local ret = C_ChatInfo.SendAddonMessage(A.CommunicationId, text, "WHISPER", request.playerName);
    --     --print("SEND IS OK: ", ret, request.playerName);
    --     --C_ChatInfo.SendAddonMessage(A.CommunicationId, text, "WHISPER", request.playerName);
    --     -- C_ChatInfo.RegisterAddonMessagePrefix(AddonPrefix);
    -- end,

    GetDungeonByRequest = function(self, request)
        local data = self:GetData();
        for _, model in pairs(data) do
            if model.checked and model.name == request.zone and model.mode == request.mode then
                return model;
                --print('>>', request.zone, request.mode);
            end
        end
    end,

    IsMyRoleSuitable = function(self, request, dungeon)
        local myrole = dungeon.role or P.role or nil;
        local iAmDd = myrole == C.Roles.mdd or C.Roles.rdd == myrole;


        if request.roles == nil or myrole == nil then
            return true
        end

        for _, rr in pairs(request.roles) do
            if rr == myrole then return true end
            if rr == C.Roles.dd and iAmDd then return true end
        end

        return false;

    end
}

local FauxFrameMixin = {
    GetData = function(self)
        return self:GetParent():GetData();
    end,
}

local ButtonConfig = {
    textures = {
        {
            parentKey = "texLeft",
            texture = "Interface\\AuctionFrame\\UI-AuctionItemNameFrame",
            point = { "LEFT", "parent", "LEFT", 2, 0 },
            width = 10,
            height = 20,
            texCoord = { 0, 0.078125, 0, 1.0 }
        },
        {
            parentKey = "texRight",
            texture = "Interface\\AuctionFrame\\UI-AuctionItemNameFrame",
            point = { "RIGHT", "parent", "RIGHT", 0, 0 },
            width = 10,
            height = 20,
            texCoord = { 0.75, 0.828125, 0, 1.0 }
        },
        {
            parentKey = "texMid",
            texture = "Interface\\AuctionFrame\\UI-AuctionItemNameFrame",
            points = {{ "LEFT", "parent", "LEFT", 12, 0 }, { "RIGHT", "parent", "RIGHT", -10, 0 }},
            width = 10,
            height = 20,
            texCoord = { 0.078125, 0.75, 0, 1.0 }
        },
    }
}

local ButtonMixin = {
    Update = function(self, data)
        self.data = data;
        data = data or {};

        local name = data.name or "";
        local mode = U:GetModeName(data.mode);
        if name ~= ""  and mode == C.Modes.Heroic[2] then
            name = name .. " - " .. mode;
        end
        self.zone:SetText(name);
        --self.mode:SetText(U:GetModeName(data.mode));

        --UIDropDownMenu_SetText(self.role, role);
        --UIDropDownMenu_SetSelectedName(self.role, role);

        self:UpdateRole();
        self:UpdateChecked();
        self:UpdateNote();
        self:UpdateColors();

    end,

    UpdateRole = function(self)

        local role = self.data and self.data.role or LOC.Empty;
        self.role:SetText(role);

    end,
    UpdateNote = function(self)
        self.note:SetText(self.data and self.data.note or "");
    end,
    UpdateColors = function(self)
        if self.data == nil then return end;
        local color = C.Colors.dungeon:GetColor(self.data);
        --C.COLORS.dungeon:GetColor(c.data);
        --MODECOLORS[self.data:IsHeroic()];
        self.zone:SetTextColor(unpack(color));
        --self.mode:SetTextColor(unpack(color));
    end,

    IsChecked = function(self)
        return self.data and self.data.checked == true;
    end,

    UpdateChecked = function(self, newValue)
        local checked;
        if newValue == nil then
            checked = self:IsChecked();
        else
            checked = newValue;
            if self.data then
                self.data.checked = newValue;
                Cache:SetOption("lfg.dungeon."..self.data.id..".checked", newValue);
            end
        end
        self.checked:SetChecked(checked);

        --if self.data == nil then return end;

        if checked and self.data ~= nil then
            self.send:Show();
            self.texLeft:Show();
            self.texMid:Show();
            self.texRight:Show();
        else
            self.send:Hide();
            self.texLeft:Hide();
            self.texMid:Hide();
            self.texRight:Hide();

        end

    end,
}

local function OnButtonChecked(self)
    local parent = self:GetParent();
    -- parent.data.checked = self:GetChecked();
    -- local skipSet = true;
    
    parent:UpdateChecked(self:GetChecked());
end

local function OnButtonEditNote(value, button)
    button.data.note = value;
    Cache:SetOption("lfg.dungeon."..button.data.id..".note", value);
    button:UpdateNote();
end

local function OnPlayerEditNote(value, button)
    P.note = value;
    Cache:SetOption("lfg.player.note", value);
    button:Update();
end

local function Add_RoleDropdownMenuItems(isChecked, onClick)

    UIDropDownMenu_AddButton({ text = LOC.Role, isTitle = true, notCheckable = true  });

    local info = {
        text = LOC.NotSpecified,
        value = "",
        checked = isChecked(nil),
        func = onClick
    };  

    UIDropDownMenu_AddButton(info);

    for _i, role in pairs(C.Roles) do
        wipe(info)
        info.text = role;
        info.value = role;
        info.checked = isChecked(role);
        info.func = onClick;
        --info.classicChecks = true;
        UIDropDownMenu_AddButton(info)
    end

    --UIDropDownMenu_AddButton({ text = "---", isTitle = true, notCheckable = true  });
    UIDropDownMenu_AddSeparator();
    UIDropDownMenu_AddButton({ text = LOC.Close, notCheckable = true });

end

local function CreateRoleDropdownMenuOptions(onClick, isChecked)
    return {
        parentKey = "dropdown",
        buildMenu = function(dropdown)
            Add_RoleDropdownMenuItems(isChecked, onClick);
        end
    }    
end

local function isRolesMeetCondition(mes_roles, my_role)
    local iAmDd = RoleApi:IsDamageDealer(my_role);
    if mes_roles == nil or my_role == nil then
        return true
    end
    for _, rr in pairs(mes_roles) do
        if rr == my_role then return true end
        if rr == RoleApi.dd.id and iAmDd then return true end
    end
    return false;
end



local FrameBuilders = {

    -- default role control
    function(f)

        

        --local fnt = h:GetFont();

        local r = CreateFrame("Button", "$parent_DefaultRoleButton", f);
        fb:AddHoverBackground(r, {0,0,0,.5});
        r:SetPoint("TOPLEFT", f, "TOPLEFT", 65, -37);
        r:SetSize(40,20);
        local t = r:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
        --U:UpdateFontString(t, nil, 7);
        t:SetAllPoints();

        t:SetTextColor(1,1,1);

        r.Update = function(self) 
            t:SetText(P.role or "---");
        end
        r:Update();
        --t:SetText(P.role or "---");

        local onClick = function(self, arg1, arg2, checked)
            local v = self.value ~= "" and self.value or nil;
            P.role = v;
            Cache:SetOption("lfg.player.role", v);
            r:Update();
        end

        local isChecked = function(role)
            return P.role == role;
        end

        local roleDropdownOptions = CreateRoleDropdownMenuOptions(onClick, isChecked);
        U:CreateDropdownMenu(r, roleDropdownOptions);

        local h = f:CreateFontString("$parent_DefaultRoleHeader", "OVERLAY", "GameFontNormalSmall");
        h:SetPoint("BOTTOM", r, "TOP", 0, 2);
        h:SetText(LOC.DefaultRole);
        --h:SetJustifyH("LEFT");
        --U:UpdateFontString(h, nil, 7);

        return r;
    end,

    -- default note control
    function(f, prev)
        
        local r = CreateFrame("Button", "$parent_DefaultNoteButton", f);
        fb:AddHoverBackground(r, {0,0,0,.5});
        r:SetPoint("LEFT", prev, "RIGHT", 20, 0);
        r:SetSize(160,20);
        

        local t = r:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
        --U:UpdateFontString(t, nil, 7);
        t:SetAllPoints();
        t:SetTextColor(1,1,1);

        r.Update = function(self) 
            t:SetText(P.note or "---");
        end
        r:Update();

        local h = f:CreateFontString("$parent_DefaultRoleHeader", "OVERLAY", "GameFontNormalSmall");
        h:SetPoint("BOTTOM", r, "TOP", 0, 2);
        h:SetText(LOC.DefaultNote);
        --h:SetJustifyH("LEFT");
        --U:UpdateFontString(h, nil, 7);

        r:SetScript("OnClick", function(self)  
            U:EditText(P.note, OnPlayerEditNote, { self }, LOC.EditNote, LOC.Change) 
        end);

    end,

    -- dungeons name filter control
    function(f)
        local anchor = f.faux.Buttons[1].zone;
        --print('--', anchor);
        local e = CreateFrame("EditBox", "$parent_Filters_Name", f, "InputBoxInstructionsTemplate");
        e:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", 5, 10);
        e:SetPoint("BOTTOMRIGHT", anchor, "TOPRIGHT", -5, 10);
        --e:SetHeight(25)
        --name:SetSize(123, 25);
        local nameHght = 20;
        e:SetHeight(nameHght);
        e.Middle:SetHeight(nameHght);
        e.Left:SetHeight(nameHght);
        e.Right:SetHeight(nameHght);
        e:SetAutoFocus(false);
        e:SetText("");
        U:UpdateFontString(e, nil, 8);
        e:SetScript("OnKeyUp", U:Debounce(function() 
            local text = e:GetText();
            if text ~= nil then
                text = string.lower(text);
            end
            if text == "" then text = nil end
            DungeonFilter:Set("name", text);
        end, .5));
        -- e:SetScript("OnEnterPressed", function() f:OnApply() end);
        -- f.edit = e;

        -- e:SetScript("OnEscapePressed", function() f:OnEsc() end);
        
    end
}


local ButtonBuilders = {

    function(b)
        local p = b:GetParent()
        if not p.lastClicked then
            p.lastClicked = GetTime() - (SEND_DELAY_IN_SECONDS);
        end
    end,

    function(b)
        local check = CreateFrame("CheckButton", "$parent_Enabled", b, "ChatConfigCheckButtonTemplate");
        b.checked = check;
        check:SetPoint("LEFT", b, "LEFT", 0, 0);
        
        check:SetHitRectInsets(0, -220, 0, 0);

        check:SetScript("OnClick", OnButtonChecked);

        return check;
    end,

    function(b, prev)
        local t = b:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
        --U:UpdateFontString(t, nil, 10);

        b.zone = t;
        t:SetSize(300, 20);
        t:SetPoint("LEFT", prev, "RIGHT", 5, 0);
        t:SetText("text")
        t:SetJustifyH("LEFT");

        return t;
    end,

    function(f, prev)

        local b = CreateFrame("Button", "$parent_SendRequest", f, "UIPanelButtonTemplate");
        f.send = b;
        
        b.enabledText = LOC.Send;
        b:SetText(b.enabledText);
        b:SetScript("OnClick", function(self) 
            local p = f:GetParent();
            p.lastClicked = GetTime();
            p:GetParent():SendRequest(f.data);
        end);
        b:SetPoint("RIGHT", prev, "RIGHT", 0, 0);
        b:SetWidth(70);
        b:SetHeight(20);
        U:UpdateFontString(b.Text, nil, 10);

        b:SetScript("OnUpdate", function(self) 
            local p = f:GetParent();
            local t = GetTime();
            local d = t - p.lastClicked;
            local remains = SEND_DELAY_IN_SECONDS - d;
            if remains > 0 then
                self:SetText(floor(remains));
                self:SetEnabled(false);
            else
                self:SetText(b.enabledText);
                self:SetEnabled(true);
            end
        end);

        return prev;
    end,

    -- function(b, prev)
    --     local t = b:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
    --     U:UpdateFontString(t, nil, 6);

    --     b.mode = t;
    --     t:SetSize(30, 20);
    --     t:SetPoint("LEFT", prev, "RIGHT", 5, 0);
    --     t:SetText("text")
    --     t:SetJustifyH("LEFT");
    --     return t;
    -- end,

    function(b, prev)
        local f = CreateFrame("Button", "$parent_RoleHld", b);
        local parent = b:GetParent();
        --local data = parent.data;

        f:SetSize(30, 20);
        f:SetPoint("LEFT", prev, "RIGHT", 5, 0);

        fb:AddHoverBackground(f, {0,0,0,.5})

        local onClick = function(self, arg1, arg2, checked)
            local v = self.value ~= "" and self.value or nil;
            b.data.role = v;
            Cache:SetOption("lfg.dungeon."..b.data.id..".role", v);
            b:UpdateRole();
        end

        local isChecked = function(role)
            return b.data and b.data.role == role;
        end

        local roleDropdownOptions = CreateRoleDropdownMenuOptions(onClick, isChecked);
        U:CreateDropdownMenu(f, roleDropdownOptions);


        local t = f:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
        --U:UpdateFontString(t, nil, 8);
        t:SetAllPoints();
        b.role = t;
        t:SetText("text")
        --t:SetJustifyH("LEFT");
        return t;
    end,

    function(b, prev)
        local f = CreateFrame("Button", "$parent_RoleHld", b);
        f:SetHeight(20);
        f:SetPoint("LEFT", prev, "RIGHT", 5, 0);
        f:SetPoint("RIGHT", b, "RIGHT", 0, 0);

        fb:AddHoverBackground(f, {0,0,0,.5})

        local t = f:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
        --U:UpdateFontString(t, nil, 7);
        t:SetAllPoints();
        b.note = t;
        t:SetText("");

        f:SetScript("OnClick", function(self)  
            --local parent = self:GetParent();
            U:EditText(b.data.note, OnButtonEditNote, { b }, LOC.EditNote, LOC.Change) 
        end);
        -- Will Be Implemented next time
        --f:Hide();

        return t;
    end,

    -- remove dropdown
    -- function(b, prev)
    --     local d = CreateFrame("Button", "$parent_RoleSelector", b, "UIDropDownMenuTemplate")
    --     d:SetPoint("LEFT", prev, "RIGHT", 5, 0);
    --     d:SetWidth(150);
    --     b.role = d;

    --     UIDropDownMenu_Initialize(d, function(self) 

    --         local click = function(self)
    --             UIDropDownMenu_SetSelectedValue(d, self.value);
    --             local val = self.value
    --             if self.value == LOC.NotSpecified then
    --                 val = nil;
    --             end
    --             b.data.role = val;
    --         end

    --         UIDropDownMenu_AddButton({
    --             text = LOC.NotSpecified,
    --             value = nil,
    --             func = click,
    --             checked = (b.data and b.data.role == nil) or not b.data;
    --         });

    --         for key, value in pairs(C.Roles) do
    --             local info = {};
    --             info.text = value;
    --             info.value = value;
    --             info.func = click;
    --             info.checked = (b.data and b.data.role == value);
    --             UIDropDownMenu_AddButton(info);
    --         end
    --     end);        
    -- end,

}



local function CreateScrollButton(index, parent, prev)
    local f = CreateFrame("Button", "$parent_Btn" .. index, parent);
    if index == 1 then
        parent.firstButton = f;
        f:SetPoint("TOPLEFT", parent, "TOPLEFT", 0, -35);
        f:SetPoint("TOPRIGHT", parent, "TOPRIGHT", -32, -35);
    else
        f:SetPoint("TOPLEFT", prev, "BOTTOMLEFT");
        f:SetPoint("TOPRIGHT", prev, "BOTTOMRIGHT");
    end

    f:SetHeight(22);
    fb:SetupTextures(f, ButtonConfig, parent);

    Mixin(f, ButtonMixin);

    fb:ApplyFrameBuilders(f, ButtonBuilders);

    return f;
end


local function AddFauxScroll(f)

    local buttonsCount = 17;

    local fauxHolder = CreateFrame("Frame", "$parent_Holder", f);
    fauxHolder:SetPoint("TOPLEFT", f, "TOPLEFT", 8, -63);
    fauxHolder:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", -3, 28);
    f.faux = fauxHolder;
    U:AddFauxScroll(fauxHolder, buttonsCount, CreateScrollButton, FauxFrameMixin);
    return fauxHolder;
end

function M:BuildFrame(parent, tabId)
    local f = CreateFrame("Frame", "$parentTab"..tabId.."Frame", parent);
    f.tabId = tabId;
    f.frameName = "LfgFrame";
    f:Hide();
    Mixin(f, FrameMixin);

    ---- add faux scroll frames

    local fauxHolder = AddFauxScroll(f);

    ---- setup update buttons handlers

    local updateButtons = function() fauxHolder:UpdateButtons(); end

    f:SetScript("OnShow", updateButtons);

    DungeonFilter:AddListener(updateButtons);

    fb:ApplyFrameBuilders(f, FrameBuilders);


    -------  chat throttle

    -- ChatThrottler.OnWipe = function(self, remains, total)
    --     print("------ ==================================== ------");
    --     print("------ # remains: ", remains, ", was: ", total);
    --     print("------ ==================================== ------");
    -- end

    --ChatThrottler:Start();
    --Bus:AddChatHandler(BaseChatHandler, 90, 60);
    --Bus:AddTypeHandler(TypeApi.Lfm.id, LfmChatHandler, nil, "LfmHandler");

    return f;
end


--### LFM HANDLER

local function LfmChatHandler(message)
    
    if not message.dungeon then return end;
    --print("--> ", message.dungeon.id);
    local dun = LFGDB:GetById(message.dungeon.id);
    if not dun or not dun.checked then 
        --print("no dungeon", dun, dun and dun.checked);
        return 
    end

    local myrole = dun.role or P.role;

    local rolesAreMatch = isRolesMeetCondition(message.roles, myrole);

    if rolesAreMatch then
        local text = P:LfmResponse(dun);
        Bus:SendAddonMessage(text, message.player);
        --print("SENDED RESPONSE");
    end

    --print("handling ?", rolesAreMatch)

end


function M:Start()
    LFGDB:Init();
    Bus:AddTypeHandler(TypeApi.Lfm.id, LfmChatHandler, nil, "LfmHandler");
end
