local A = select(2, ...);
local LFG = A.Lfg;
local M = A.Lfm;
local fb = A.Lib.FB;
local C = A.Constants;
local D = A.Api.Dungeon;
local L = A.Localize.translate;
local U = A.Utils;
local P = A.Player;
local R = A.Request;
local Bus = A.ChatBus;
local MSG = A.Message;
local Companion = A.Companion;
local TypeApi = A.Api.Type;
local RoleApi = A.Api.Role;
local DungeonApi = A.Api.Dungeon;

local HUNDRED_SECONDS = 100;
local HUNDRED_MINUTES = 100 * 60;
local SECONDS_IN_HOUR = 60 * 60;

-- local MODECOLORS = {
--     [true] = "ff6000", -- { 1, .4, 0 },
--     [false] = "c0ffc0", --{ .8, 1, .8 },
-- }

local MODECOLORSTABLE = {    
    [true] = { 1, .5, 0 },
    [false] = { .6, 1, .5 }
}

local SEND_DELAY_IN_SECONDS = 20;
local REQ = {};
local THIRTY_MINUTES = 60 * 30;
local OLD_WHEN = 60 * 5;

StaticPopupDialogs["TabuLfg.Lfm.SendInvite"] = {
    Ask = function(self, userName, OnAccept)
        self.OnAccept = OnAccept;
        self.text = L("Do you want to invite " .. userName .. "?");
        StaticPopup_Show("TabuLfg.Lfm.SendInvite");
    end,
    button1 = L("Yes, invite"),
    button2 = L("Cancel"),
    timeout = 0,
    whileDead = true,
    hideOnEscape = true,
    preferredIndex = 3,  -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
}

local AskAboutInvite = StaticPopupDialogs["TabuLfg.Lfm.SendInvite"];


local function InvokeHandlers(handlers, ...)
    for _, handler in pairs(handlers) do
        handler(...);
    end
end

local function LfmUserSorter(a,b)
    local ad = a.dungeon;
    local bd = b.dungeon;
    local an = (a.player or {}).fullName;
    local bn = (b.player or {}).fullName;
    if ad == nil and bd == nil then
        return an < bn;
    elseif ad == nil and bd ~= nil then
        return false;
    elseif ad ~= nil and bd == nil then
        return true;
    elseif ad.id == bd.id then
        return an < bn;
    else
        return ad.id < bd.id
    end
end

local DATA = {

    users = {}, 
    usersByName = {},

    GetUser = function(self, name)
        local users = self:GetUsersByName();
        return users[name];
    end,

    SetUser = function(self, user)
        local users = self:GetUsersByName();
        users[user.player.fullName] = user;
    end,

    RemoveUser = function(self, user)
        local users = self:GetUsersByName();
        users[user.player.fullName] = nil;        
    end,

    GetUsersByName = function(self)
        if not self.usersByName then
            self.usersByName = {}
        end
        return self.usersByName;
    end,

    EntryIsFreshEnough = function(self, user, now)
        now = now or GetTime();
        return now - user.time < OLD_WHEN;
    end,

    EntryIsMatch = function(self, user, hasZone)
        hasZone = hasZone or REQ:HasZone();
        if hasZone then
            return user.dungeon and REQ:IsZoneIdEqual(user.dungeon.id) or false;
        else
            return MSG:IsMatch(user, REQ.weakFilter or {});
        end
        if hasZone and (not user.dungeon or not REQ:IsZoneIdEqual(user.dungeon.id)) then 
            return false 
        end
        --if not hasZone and user.dungeon then return false end
        return true;
    end,

    dataUpdateHandlers = {},
    

    OnDataUpdate = function (self, res)
        InvokeHandlers(self.dataUpdateHandlers, res);
        -- for _, handler in pairs(self.dataUpdateHandlers) do
        --     handler(res);
        -- end
    end,

    AddOnDataUpdateHandler = function(self, handler)
        table.insert(self.dataUpdateHandlers, handler);
    end,

    _Get_ProcessOne = function(self, entry, result, options)
        if not entry.id or options.keys[entry.id] then return end
        options.keys[entry.id] = true;


        if not self:EntryIsFreshEnough(entry, now) then return end
        if not self:EntryIsMatch(entry, options.hasZone) then return end
        
        table.insert(result, entry);
        
    end,

    Get = function(self)
        local res = {};
        local keys = {};
        local now = GetTime();
        local hasZone = REQ:HasZone();
        local users = self:GetUsersByName();
        local options = {
            now = now,
            hasZone = hasZone,
            keys = keys,
        }
        local index = 0;
        for _, user in pairs(users) do
            index = index + 1;
            options.index = index;

            self:_Get_ProcessOne(user, res, options);

        end

        table.sort(res, LfmUserSorter)

        self:OnDataUpdate(res);
        return res;
    end,

    AddPlayer = function(self, user)
        local id = user.id
        
        local entry = self:GetUser(id)

        if not entry then
            entry = user;
            self:SetUser(entry);
        else
            Mixin(entry, user)
        end

        self:OnPlayerUpdate();

        --print("< player updated >");

    end,

    playerUpdateHandlers = {},

    AddOnPlayerUpdateHandler = function(self, handler)
        table.insert(self.playerUpdateHandlers, handler);
    end,

    OnPlayerUpdate = function(self, player)
        InvokeHandlers(self.playerUpdateHandlers, player);
    end,

    SetDungeon = function(self, id)
        if self.currentId ~= id then
            self.currentId = id;
            self:CleanUp();
        end
    end,

    CleanUp = function(self)
        wipe(self.users); -- = {};
    end,
}

DATA._removeOldEntriesTimer = C_Timer.NewTicker(60, function()  
    local users = DATA:GetUsersByName();

    for key, user in pairs(users) do
        if not DATA:EntryIsFreshEnough(user, now) then
            DATA:RemoveUser(user);
        end
    end

end);

DATA:AddOnDataUpdateHandler(function(result) 
    if REQ:HasZone() then
        Companion:Update(result);
    end
end)

Bus:AddTypeHandler(nil, function(message) 
    DATA:AddPlayer(message);
end, nil, "All handler");


Mixin(REQ,{
    updateHandlers = {},
    weakFilterHandlers = {},
    values = {},
    weakFilter = {
        eras = {},
        modes = {},        
    },

    OnFilterChange = function(self)
        for _, handler in pairs(self.weakFilterHandlers) do
            handler();
        end
    end,
    AddOnFilterChange = function(self, handler)
        table.insert(self.weakFilterHandlers, handler);
    end,

    OnUpdate = function(self)
        for _, handler in pairs(self.updateHandlers) do
            handler();
        end
    end,

    AddOnUpdate = function(self, handler)
        table.insert(self.updateHandlers, handler);
    end,



    Set = function(self, arg1, arg2)
        local t = type(arg1);
        if t == "table" then
            if arg2 == true then
                wipe(self.values)
            end
            Mixin(self.values, arg1);
        elseif t == "string" then
            self.values[arg1] = arg2;
        end
        self:OnUpdate();
        --print(": set", arg1, arg2)
    end,

    Get = function(self, key)
        if key == nil then
            return self.values;
        else
            return self.values[key];
        end
    end,

    SetZoneId = function(self, id)
        self:Set("zoneId", id);
        local zone = id ~= nil and D:GetOne(id) or nil;
        self:Set("zone", zone);
    end,

    GetZoneId = function(self, id)
        return self:Get("zoneId");
    end,

    GetZone = function(self)
        return self:Get("zone");
    end,

    HasZone = function(self)
        return self:GetZone() ~= nil;
    end,

    IsZoneIdEqual = function(self, id)
        return self:GetZoneId() == id;
    end,

    GetNote = function(self, ifEmpty)
        return self:Get("note") or ifEmpty;
    end,

    SetNote = function(self, value)
        self:Set("note", value);
    end,

    GetRolesText = function(self, ifEmpty)

        local roles = self:Get("roles");
        if not roles then
            return ifEmpty
        end

        local t = {};
        for role, roleValue in pairs(roles) do
            if roleValue == true then
                table.insert(t, role);
            end
        end
        local res = U:JoinToString(t, C.RequestRoleSeparator);
        if res == nil or res == "" then
            res = ifEmpty;
        end
        return res;
    end,

    SetRole = function(self, role, val)
        local roles = self:Get("roles") or {};
        roles[role] = (val == true);
        self:Set("roles", roles);
    end,

    HasRole = function(self, role)
        local roles = self:Get("roles");
        return roles ~= nil and roles[role] == true;
    end,
    RoleIsMatch = function(self, role)
        local roles = self:Get("roles");
        return roles == nil or roles[role] == true;
    end,
    -----------
    lastSendedRequest = GetTime()  - (SEND_DELAY_IN_SECONDS),

    UpdateLastSendedRequest = function(self)
        self.lastSendedRequest = GetTime();
    end,

    GetSecondsToNewRequest = function(self)
        local t = GetTime();
        local d = t - self.lastSendedRequest;
        return SEND_DELAY_IN_SECONDS - d;
    end,
    ------------------------

    SendRequest = function(self)

        self:UpdateLastSendedRequest();

        local text = self:GetRequestText();
        if text == nil then
            --print('Request was not sent, dungeon not found');
            return;
        end

        Bus:SendChatMessage(text);

    end,
    GetRequestText = function(self)
        local dun = self:GetZone();
        if dun == nil then return end;
        local t = {
            C.Lfm, 
            dun.name  .. C.RequestModeSeparator .. dun.mode,
            self:GetRolesText(),
            self:GetNote()
        }
        return U:JoinToString(t, C.RequestSeparator);
    end,
});

REQ:AddOnUpdate(function() Companion:Update() end)

local LOC = {
    SelectDungeon = L("Select dungeon"),
    NotSpecified = L("Not Specified"),
    Close = L("Close"),
    Dungeons = L("Dungeons"),
    Mode = L("Mode"),
    Era = L("Addon"),
    Send = L("Send"),
    Invite = L("Invite"),
    Whisper = L("Whisper"),
    ClearZone = L("Unselect dungeon"),
    DungeonNotSelected = L("Dungeon not selected"),
    DungeonNote = L("Options"),
    EditNote = L("Edit dungeon options"),
    Change = L("Change"),
    Roles = L("Roles")
}

local DungeonFilter = {
    era = C.Eras.BC,
    heroic = false
}


local FauxFrameMixin = {
    -- UpdateData = function(self)

    -- end,
    -- GetData = function(self)
    --     return DATA:Get();
    -- end,
}

local ButtonConfig = {
    textures = {
        {
            parentKey = "texLeft",
            texture = "Interface\\AuctionFrame\\UI-AuctionItemNameFrame",
            point = { "LEFT", "parent", "LEFT", 2, 0 },
            width = 10,
            height = 25,
            texCoord = { 0, 0.078125, 0, 1.0 }
        },
        {
            parentKey = "texRight",
            texture = "Interface\\AuctionFrame\\UI-AuctionItemNameFrame",
            point = { "RIGHT", "parent", "RIGHT", 0, 0 },
            width = 10,
            height = 25,
            texCoord = { 0.75, 0.828125, 0, 1.0 }
        },
        {
            parentKey = "texMid",
            texture = "Interface\\AuctionFrame\\UI-AuctionItemNameFrame",
            points = {{ "LEFT", "parent", "LEFT", 12, 0 }, { "RIGHT", "parent", "RIGHT", -10, 0 }},
            width = 10,
            height = 25,
            texCoord = { 0.078125, 0.75, 0, 1.0 }
        },
        {
            parentKey = "icon",
            layer = "BORDER",
            point = { "LEFT" },
            width = 32,
            height = 32,
        },
        {
            parentKey = "iconBorder",
            texture = "Interface\\Common\\WhiteIconFrame",
            point = { "LEFT" },
            width = 32,
            height = 32,
            hidden = true,
        },
    }
}

local ButtonMixin = {
    Update = function(self, data)

        self.data = data;

        -- Old
        -- self.name:Update();
        -- self.note:Update();
        -- self:UpdateColor();

        -- New
        for _, item in pairs(self.items) do
            if type(item) == "function" then
                item();
            elseif item.Update then
                item:Update();
            end
        end

        if self.data and self.data:IsDetected() then
            --self.typeIcon:Show();
            self.dungeon:Show();
            self.roles:Show();
            self.note:ReAttach(self.roles);
        else
            --self.typeIcon:Hide();
            self.dungeon:Hide();
            self.roles:Hide();
            self.note:ReAttach(self.typeIcon);
        end

        --self.dungeon:Update();
    end,

    UpdateRole = function(self)
        -- if self.data.role then
        --     self.role:SetText(self.data.role);
        -- elseif self.data.roles then
        --     local txt = U:JoinToString(self.data.roles, ", ");
        --     self.role:SetText(txt);
        -- else
        --     self.role:SetText("");
        -- end
    end,

    UpdateColor = function(self)
        -- local r, g, b = GetClassColor(self.data.player.class);
        -- self.name:SetTextColor(r,g,b);

        -- if self.data.native then
        --     r,g,b = 0,1,.3;
        -- else
        --     r,g,b = 1,1,1;
        -- end
        -- self.note:SetTextColor(r,g,b);
        --self.class:SetTextColor(r,g,b);
    end,

    UpdateTime = function(self)

        if not self.data or not self.data.time then
            self.time:SetText("");    
        end

        local text = "";
        local secondsAgo = floor(GetTime() - self.data.time);

        if secondsAgo < 100 then
            text = "-" .. secondsAgo .. "s"
        elseif secondsAgo < HUNDRED_MINUTES then
            local minutesAgo = ceil(secondsAgo / 60);
            text = "-" .. minutesAgo .. "m";
        else
            local hoursAgo = ceil(secondsAgo / SECONDS_IN_HOUR);
            text = "-" .. hoursAgo .. "m";
        end

        self.time:SetText(text);
    end,

    OnFrameUpdate = function(self, elapsed)
        self._lastFrameUpdate = self._lastFrameUpdate + elapsed;
        if self._lastFrameUpdate <= self._frameUpdateInterval then return end;

        self:UpdateTime();

        self._lastFrameUpdate = 0;
    end,

    PopulateTooltip = function(self)
        
        GameTooltip:SetOwner(self);
        if not self.data.native then
            GameTooltip:AddLine(self.data.text, nil, nil, nil, true);
            if self.data.undetected_words and #self.data.undetected_words > 0 then
                local words = U:JoinToString(self.data.undetected_words, ", ");
                GameTooltip:AddLine("---undetected---");
                GameTooltip:AddLine(words, nil, nil, nil, true);
            end
            -- GameTooltip:AddLine("---");
            -- GameTooltip:AddLine(self.data.normalizedText);
        else
            --GameTooltip:AddLine("Tabu-Lfg auto response:");
            local pl = self.data.player;
            local role = self.data.role and self.data.role.name or "";

            U:InlineColor()
            local player_text = pl.name or "";
            local clscolor; 
            if pl.class then
                local _,_,_, rgb = GetClassColor(pl.class);
                clscolor = rgb;
            end
            if pl.level then
                player_text = pl.level .. " " .. player_text;                
            end

            GameTooltip:AddDoubleLine(U:InlineColor(player_text, clscolor), U:InlineColor(role, "50ff30"))
            GameTooltip:AddLine("Note: " .. (self.data.note or ""));
            GameTooltip:AddDoubleLine("Item level:", self.data.avgILevel);

        end
        GameTooltip:AddLine("---");
        GameTooltip:AddLine(U:InlineColor(L("Left Click"),"0080ff") .. " - " .. U:InlineColor(L("Send whisper to player"),"ffffff"));
        GameTooltip:AddLine(U:InlineColor(L("Right Click"),"0080ff") .. " - " .. U:InlineColor(L("Send invite to player"),"ffffff"));

    end,

    OnMouseEnter = function(self)
        self:PopulateTooltip();
        GameTooltip:Show();
    end,


    OnMouseLeave = function(self)
        GameTooltip:Hide();
    end,

    OnClick = function(self, button)
        local isShift = IsShiftKeyDown();
        local isAlt = IsAltKeyDown();
        if not isShift and button == "LeftButton" then
            ChatFrame_OpenChat("/w " .. self.data.player.name .. " ");
            return;
        end

        if not isShift and button == "RightButton" then
            local userName = self.data.player.name;
            AskAboutInvite:Ask(userName, function() 
                InviteUnit(userName);
            end);
            return;
        end


        if isShift and isAlt and button == "LeftButton" and Tabu then
            Tabu.dump(self.data);
        end
    end,
}



local function RoleIconTexture_Update(self, roles)
    if not roles or #roles == 0 or not RoleApi:HasRole(roles, self.roleId) then
        self:SetTexture(nil);
        -- self:SetWidth(0);
        -- return;
    else
        --self:SetWidth(self.sizePx);
        self:SetAtlas(self.atlasId)
    end
end

local function CreateRoleIconTexture(parent, prev, roleId, atlasId)
    local t = parent:CreateTexture(nil, "OVERLAY");
    if prev == nil then
        t:SetPoint("LEFT", parent, "LEFT", 0, 0);
    else
        t:SetPoint("LEFT", prev, "RIGHT", 0, 0);
    end
    t.sizePx = 20;
    t.roleId = roleId;
    t.atlasId = atlasId; --"groupfinder-icon-role-large-tank";
    t:SetSize(t.sizePx, t.sizePx);
    t.Update = RoleIconTexture_Update;
    return t;
end


local ButtonBuilders = {

    function(b)
        b.items = {};
    end,

    -- time
    function(b)

        local t = b:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
        U:UpdateFontString(t, nil, 8);
        t:SetPoint("LEFT", b, "LEFT", 5, 0);
        t:SetWidth(30);
        b.time = t;
        return t;

    end,

    -- function(b, prev)
    --     local t = CreateRoleIconTexture(b, prev);
    --     t.Update = function(self)
    --         local role = b.data and b.data.role or nil;
    --         if not role then
    --             print(" > role missing");
    --             self:SetTexture(nil);
    --             self:SetWidth(0);
    --         else
    --             print(" > role exist");
    --             self:SetWidth(self.sizePx);
    --             self:SetAtlas(role.iconAtlas);
    --         end
    --     end
    --     table.insert(b.items, t);
    --     return t;
    -- end,

    -- name
    function(b, prev)
        local t = b:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");

        b.name = t;
        t.originalWidth = 90;
        t:SetSize(t.originalWidth, 20);
        t:SetPoint("LEFT", prev, "RIGHT", 5, 0);
        t:SetJustifyH("LEFT");

        t.Update = function(self)

            if not b.data then
                self:SetText("")
                return;
            end

            self:SetTextColor(GetClassColor(b.data.player.class));

            local name = b.data.player.name or "";
            local level = b.data.player.level;
            level = level and (level .. " ") or "";

            self:SetText(level .. name);
        end
        table.insert(b.items, t);
        return t;
    end,   
    
    -- LFM or LFG Icon 
    -- .typeIcon
    function(b, prev)
        local t = b:CreateTexture(nil, "OVERLAY");
        t:SetPoint("LEFT", prev, "RIGHT", 5, 0);
        b.typeIcon = t;

        t:SetSize(20, 20);
        
        t.Update = function(self)
            local typeId = b.data.type and b.data.type.id or nil;
            local roleIcon = b.data.role and b.data.role.iconAtlas or nil;

            if typeId == TypeApi.Lfm.id then
                t:SetAtlas("socialqueuing-icon-group");
            elseif typeId == TypeApi.Lfg.id then
                t:SetAtlas(roleIcon or "groupfinder-icon-friend");
            elseif roleIcon then
                t:SetAtlas(roleIcon);
            else
                self:SetTexture(nil);
            end
            
        end

        table.insert(b.items, t);
        return t;
    end,

    -- Dungeon/quest name
    function(b, prev)
        local t = b:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
        b.dungeon = t;
        --U:UpdateFontString(t, nil, 8);
        t:SetPoint("LEFT", prev, "RIGHT", 5, 0);
        t:SetWidth(150);
        t:SetJustifyH("LEFT");
        t.Update = function(self)
            local dun = b.data.dungeon;
            local qst = b.data.quest;

            if dun then
                local dungeonColor = C.Colors.dungeon:GetColor(dun); --MODECOLORSTABLE[dun:IsHeroic()];
                self:SetText(dun:ToString());
                self:SetTextColor(unpack(dungeonColor));
            elseif qst then
                self:SetText(qst.name);
                self:SetTextColor(1,1,0);
            else
                self:SetText("");
                self:SetTextColor(1,1,1);
            end

        end

        table.insert(b.items, t);
        return t;
    end,

    -- Roles
    function(b, prev)
        local f = CreateFrame("Frame", nil, b);
        f:SetSize(60, 20);
        f:SetPoint("LEFT", prev, "RIGHT", 5, 0);
        
        b.roles = f;

        local tank = CreateRoleIconTexture(f, nil, RoleApi.tank.id, "groupfinder-icon-role-large-tank");
        local heal = CreateRoleIconTexture(f, tank, RoleApi.heal.id, "groupfinder-icon-role-large-heal");
        local dd = CreateRoleIconTexture(f, heal, { RoleApi.dd.id, RoleApi.rdd.id, RoleApi.mdd.id,  }, "groupfinder-icon-role-large-dps");

        f.Update = function()
            local roles = not b.data.role and b.data.roles or nil;
            tank:Update(roles);
            heal:Update(roles);
            dd:Update(roles);
        end

        table.insert(b.items, f);
        return f;
    end,

    -- note
    function(b, prev)
        local t = b:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
        t:SetTextColor(1,1,1);
        b.note = t;
        t:SetPoint("LEFT", prev, "RIGHT", 5, 0);
        t:SetPoint("RIGHT", b, "RIGHT", -5, 0);
        t:SetHeight(15);
        t:SetJustifyH("LEFT");

        t.Update = function(self)
            if not b.data then
                self:SetText("")
                return;
            end

            local noteText = b.data.note or b.data.autoNote;
            --local text = noteText or b.data.text;


            if b.data:IsDetected() then
                if b.data.quest then
                    self:SetText(b.data.quest.text);
                else
                    self:SetText(noteText);
                end
            else
                self:SetText(b.data.text);
            end
        end

        t.ReAttach = function(self, prevEl)
            t:ClearAllPoints();
            t:SetPoint("LEFT", prevEl, "RIGHT", 5, 0);
            t:SetPoint("RIGHT", b, "RIGHT", -5, 0);
        end

        table.insert(b.items, t);
        return t;
    end,  
}


local function OnZoneEditNote(value, frame)
    REQ:SetNote(value);
end

local function DungeonToMenuItem(dun, click, isChecked)
    local info = {
        text = dun.name,
        value = dun.id,
        func = click,
        checked = isChecked(dun.id),
    };
    return info;
end

local function GetDungeonsMenu(filters, click, isChecked)
    local dungeons = D:GetAll(filters)
    local result = {};
    for _, dungeon in pairs(dungeons) do
        table.insert(result, DungeonToMenuItem(dungeon, click, isChecked))
    end
    return result;
end

local function GetZoneMenuList(click, isChecked)
    
    local menuList = {
        { text = LOC.ClearZone, value = "", func = click, checked = isChecked(nil) },
        { separator = true },
        { text = LOC.SelectDungeon, isTitle = true, notCheckable = true  },
        { 
            text = L("Classic dungeons"), hasArrow = true, 
            menuList = GetDungeonsMenu({ era = C.Eras.Classic, raid = false }, click, isChecked)
        },
        { 
            text = L("Classic raids"), hasArrow = true, 
            menuList = GetDungeonsMenu({ era = C.Eras.Classic, raid = true }, click, isChecked)
        },
        {
            text = L("BC normal dungeons"), hasArrow = true,
            menuList = GetDungeonsMenu({ era = C.Eras.BC, heroic = false, raid = false }, click, isChecked)
        },
        {
            text = L("BC heroic dungeons"), hasArrow = true,
            menuList = GetDungeonsMenu({ era = C.Eras.BC, heroic = true, raid = false }, click, isChecked)
        },
        {
            text = L("BC raids"), hasArrow = true,
            menuList = GetDungeonsMenu({ era = C.Eras.BC, raid = true }, click, isChecked)
        },
        { separator = true },
        { text = LOC.Close, notCheckable = true },
    }
    return menuList;
end

local function GetWeakFilterDropdownMenuList()

    local menu = {}
    local click, checked;

    table.insert(menu, { text = L("Show dungeons"), notCheckable = true, isTitle = true });

    click = function(self, arg1, arg2, checked) 
        local val = REQ.weakFilter.eras[self.value];
        if val == nil then val = true end
        REQ.weakFilter.eras[self.value] = not val;
        REQ:OnFilterChange();
    end

    checked = function(self)
        return REQ.weakFilter.eras[self.value] ~= false;
    end

    for key, eraName in pairs(C.Eras) do
        table.insert(menu, { text = eraName, value = eraName, func = click, checked = checked, keepShownOnClick = true });
    end

    --table.insert(menu, { separator = true });
    --table.insert(menu, { text = "Dungeon mode", notCheckable = true, isTitle = true });

    click = function(self, arg1, arg2, checked) 
        local val = REQ.weakFilter.modes[self.value];
        if val == nil then val = true end
        REQ.weakFilter.modes[self.value] = not val;
        REQ:OnFilterChange();
    end

    checked = function(self)
        return REQ.weakFilter.modes[self.value] ~= false;
    end

    for key, mode in pairs(C.Modes) do
        table.insert(menu, { text = mode[2], value = mode[1], func = click, checked = checked, keepShownOnClick = true });
    end
    table.insert(menu, { text = L("From my LFG list only"), value = true, keepShownOnClick = true, func = function(self, a1,a2, checked) 
        local val = REQ.weakFilter.inMyWishList;
        if val == nil then val = false end
        REQ.weakFilter.inMyWishList = not val;
        REQ:OnFilterChange();
    end, checked = function(self) 
        return REQ.weakFilter.inMyWishList == true;
    end });

    table.insert(menu, { separator = true });

    table.insert(menu, { text = L("Show quests"), value = true, keepShownOnClick = true, func = function(self, a1,a2,checked) 
        local val = REQ.weakFilter.showQuests;
        if val == nil then val = true end
        REQ.weakFilter.showQuests = not val;
        REQ:OnFilterChange();
    end, checked = function(self) return REQ.weakFilter.showQuests ~= false   end });

    table.insert(menu, { text = L("Show undetected"), value = true, keepShownOnClick = true, func = function(self, a1,a2,checked) 
        local val = REQ.weakFilter.showUndetected;
        if val == nil then val = true end
        REQ.weakFilter.showUndetected = not val;
        REQ:OnFilterChange();
    end, checked = function(self) return REQ.weakFilter.showUndetected ~= false   end });

    table.insert(menu, { separator = true });
    table.insert(menu, { text = LOC.Close, notCheckable = true });

    return menu;
end

local function CreateArrowButton(globalName, parent, point, size)
    size = size or 24;
    local btn = CreateFrame("Button", globalName, parent);
    btn:SetSize(size,size);

    btn:SetPoint(unpack(point));

    local nt = btn:CreateTexture(nil, "OVERLAY");
    nt:SetTexture([[Interface\ChatFrame\UI-ChatIcon-ScrollDown-Up]]);
    nt:SetSize(size,size);
    nt:SetPoint("RIGHT");
    btn:SetNormalTexture(nt);

    local pt = btn:CreateTexture(nil, "OVERLAY");
    pt:SetTexture([[Interface\ChatFrame\UI-ChatIcon-ScrollDown-Down]]);
    pt:SetSize(size,size);
    pt:SetPoint("RIGHT");
    btn:SetPushedTexture(pt);

    local dt = btn:CreateTexture(nil, "OVERLAY");
    dt:SetTexture([[Interface\ChatFrame\UI-ChatIcon-ScrollDown-Disabled]]);
    dt:SetSize(size,size);
    dt:SetPoint("RIGHT");
    btn:SetDisabledTexture(dt);

    local ht = btn:CreateTexture(nil, "OVERLAY");
    ht:SetTexture([[Interface\Buttons\UI-Common-MouseHilight]]);
    ht:SetBlendMode("ADD");
    ht:SetSize(size,size);
    ht:SetPoint("RIGHT");
    btn:SetHighlightTexture(ht);
    return btn;
end

local Builders = {

    -- select dungeon control
    function(f)



        --local fnt = h:GetFont();

        local r = CreateFrame("Button", "$parent_ZoneButton", f);
        fb:AddHoverBackground(r, {0,0,0,.5});
        r:SetPoint("TOPLEFT", f, "TOPLEFT", 55, -37);
        r:SetSize(200, 20);

        local t = r:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
        --U:UpdateFontString(t, nil, 7);
        t:SetPoint("LEFT", r, "LEFT", 20, 0);
        t:SetPoint("RIGHT", r, "RIGHT", 0, 0);
        t:SetJustifyH("LEFT");
        t:SetTextColor(1,1,1);

        r.Update = function(self) 
            local text = LOC.DungeonNotSelected;
            local zone = REQ:GetZone();
            if zone then 
                text = zone:ToString();
            end;
            t:SetText(text)
        end
        r:Update();

        f.zone = r;


        local h = f:CreateFontString("$parent_DefaultRoleHeader", "OVERLAY", "GameFontNormalSmall");
        h:SetPoint("BOTTOMLEFT", r, "TOPLEFT", 10, 2);
        h:SetText(LOC.SelectDungeon);
        h:SetJustifyH("LEFT");
        --U:UpdateFontString(h, nil, 7);

        local arrowBtn = CreateArrowButton("$parent_ArrowButton", r, { "LEFT", r, "LEFT", 0, 0 }, 16)

        local autoHideDelay, x, y, displayMode = 0, 20, 20, "MENU";
        local dropdown = CreateFrame("Frame", "$parent_ZoneDropdown", r, "UIDropDownMenuTemplate");

        local click = function(self) 
            local value = self.value;
            if value == "" then value = nil end
            REQ:SetZoneId(value);
            --DATA:SetDungeon(self.value);
            f.faux:UpdateButtons();
            r:Update();
            CloseDropDownMenus();
        end

        local isChecked = function(id) 
            return REQ:IsZoneIdEqual(id); 
        end

        local openDropdown = function() 
            PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);
            local menuList = GetZoneMenuList(click, isChecked);
            U:EasyMenu(menuList, dropdown, r, 0, 0, displayMode);
        end

        r:SetScript("OnClick", openDropdown);
        arrowBtn:SetScript("OnClick", openDropdown);

        return r;
    end,

    function(f, prev)
        
        local r = CreateFrame("Button", "$parent_DefaultNoteButton", f);
        fb:AddHoverBackground(r, {0,0,0,.5});
        r:SetPoint("LEFT", prev, "RIGHT", 10, 0);
        r:SetSize(120,20);
        

        local t = r:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
        --U:UpdateFontString(t, nil, 7);
        t:SetAllPoints();
        t:SetTextColor(1,1,1);
        f.note = r;
        r.Update = function(self) 
            t:SetText(REQ:GetNote("---"));
        end
        r:Update();

        local h = f:CreateFontString("$parent_DefaultRoleHeader", "OVERLAY", "GameFontNormalSmall");
        h:SetPoint("BOTTOM", r, "TOP", 0, 2);
        h:SetText(LOC.DungeonNote);
        --h:SetJustifyH("LEFT");
        --U:UpdateFontString(h, nil, 7);

        r:SetScript("OnClick", function(self)  
            U:EditText(REQ:GetNote(), function(value) REQ:SetNote(value); r:Update() end, nil, LOC.EditNote, LOC.Change) 
        end);

        return r;
    end,

    function(f, prev)

        local r = CreateFrame("Button", "$parent_RoleButton", f);
        fb:AddHoverBackground(r, {0,0,0,.5});
        r:SetPoint("LEFT", prev, "RIGHT", 10, 0);
        r:SetSize(120,20);

        local t = r:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
        --U:UpdateFontString(t, nil, 7);
        t:SetAllPoints();

        t:SetTextColor(1,1,1);

        r.Update = function(self) 
            t:SetText(REQ:GetRolesText(" --- "));
        end
        r:Update();

        local click = function(self, arg1, arg2, checked)
            REQ:SetRole(self.value, checked);
            r:Update();
        end

        local isChecked = function(role)
            return REQ:HasRole(role);
        end


        local dropdown = CreateFrame("Frame", "$parent_RolesDropdown", r, "UIDropDownMenuTemplate");
        local h = f:CreateFontString("$parent_RolesHeader", "OVERLAY", "GameFontNormalSmall");
        h:SetPoint("BOTTOM", r, "TOP", 0, 2);
        h:SetText(LOC.Roles);

        --U:UpdateFontString(h, nil, 7);        


        local openDropdown = function() 
            PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);

            local menuList = {
                { text = LOC.Roles, isTitle = true, notCheckable = true  }
            };

            for _i, role in pairs(C.Roles) do
                table.insert(menuList, {
                    text = role,
                    value = role,
                    checked = isChecked(role),
                    func = click,
                    keepShownOnClick = true
                });        
            end


            table.insert(menuList, { separator = true });
            table.insert(menuList, { text = LOC.Close, notCheckable = true });

            U:EasyMenu(menuList, dropdown, r, 0, 0, "MENU");
        end

        r:SetScript("OnClick", openDropdown);

        return r;

    end,

    -- SendRequest button
    function(f, prev)
        local lfgb = CreateFrame("Button", "$parent_SendRequest", f, "UIPanelButtonTemplate");
        --lfgb.lastClicked = GetTime() - (SEND_DELAY_IN_SECONDS - 10);
        lfgb.enabledText = LOC.Send;
        lfgb:SetText(lfgb.enabledText);
        lfgb:SetScript("OnClick", function(self) 
            REQ:SendRequest();
        end);
        lfgb:SetPoint("LEFT", prev, "RIGHT", 5, 2);
        lfgb:SetWidth(70);
        --local fs = lfgb.Text:GetFont()
        --U:UpdateFontString(lfgb.Text, nil, 7);

        lfgb:SetScript("OnUpdate", function(self) 
            local remains = REQ:GetSecondsToNewRequest();
            if remains > 0 then
                self:SetText(floor(remains));
                self:SetEnabled(false);
                return;
            else
                self:SetText(lfgb.enabledText);
            end
            local hasId = REQ:GetZoneId() ~= nil;
            self:SetEnabled(hasId);
        end);

        f.send = lfgb;
    end,

    -- weak filter
    function(f)
        local btn = f.faux.Buttons[1];
        local b = CreateFrame("Button", "$parent_WeakFilter", f);
        b:SetPoint("BOTTOMLEFT", btn, "TOPLEFT", 5, 3);
        b:SetSize(100, 20);
        fb:AddHoverBackground(b, {0,0,0,.5});
        local arrowBtn = CreateArrowButton("$parent_ArrowButton", b, { "LEFT", b, "LEFT", 0, 0 }, 16);

        local t = b:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
        --t:SetAllPoints();
        t:SetPoint("LEFT", arrowBtn, "RIGHT", 5, 0);

        t:SetText(L("Setup filters"));
        t:SetJustifyH("LEFT");
        --print("# btn", btn);
        local dropdown = CreateFrame("Frame", "$parent_Dropdown", b, "UIDropDownMenuTemplate");
        b:SetScript("OnClick", function() 
            PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);
            local menu = GetWeakFilterDropdownMenuList();
            U:EasyMenu(menu, dropdown, b, 0, 0, "MENU");
            --EasyMenu()
        end);

        return b;
    end,
}


local function CreateScrollButton(index, parent, prev)
    local f = CreateFrame("Button", "$parent_Btn" .. index, parent);
    if index == 1 then
        f:SetPoint("TOPLEFT", parent, "TOPLEFT", 0, -27);
        f:SetPoint("TOPRIGHT", parent, "TOPRIGHT", -32, -27);
    else
        f:SetPoint("TOPLEFT", prev, "BOTTOMLEFT");
        f:SetPoint("TOPRIGHT", prev, "BOTTOMRIGHT");
    end
    f:RegisterForClicks("AnyUp");
    f:SetHeight(27);
    fb:SetupTextures(f, ButtonConfig, parent);

    Mixin(f, ButtonMixin);

    fb:ApplyFrameBuilders(f, ButtonBuilders);
    f:SetScript("OnUpdate", f.OnFrameUpdate);
    f:SetScript("OnShow", function(self) self._lastFrameUpdate = 0; self._frameUpdateInterval = 0.5; end);
    f:SetScript("OnEnter", f.OnMouseEnter);
    f:SetScript("OnLeave", f.OnMouseLeave);


    f:SetScript("OnClick", f.OnClick);

    return f;
end



function M:BuildFrame(parent, tabId)
    local f = CreateFrame("Frame", "$parentTab"..tabId.."Frame", parent);
    f:Hide();
    f.tabId = tabId;
    f.frameName = "LfmFrame";
    
    --Mixin(f, FrameMixin);

    

    local fauxHolder = CreateFrame("Frame", "$parent_Holder", f);
    f.faux = fauxHolder;
    fauxHolder:SetPoint("TOPLEFT", f, "TOPLEFT", 8, -63);
    fauxHolder:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", -3, 28);
    local buttonsCount = 14;

    U:AddFauxScroll(fauxHolder, buttonsCount, CreateScrollButton, {

        GetData = function(self)
            local d = DATA:Get();
            return d;
        end,

        UpdateButtonByItem = function(self, item)
            if not item or not item.__button_index then return end;
            local btn = self:GetButton(item.__button_index);
            if not btn or btn.data ~= item then return end;

            btn:Update(item);

        end,
    });

    fb:ApplyFrameBuilders(f, Builders);

    f:SetScript("OnShow", function(self) 
        -- DATA.OnAdd = function() f.faux:UpdateButtons() end;
        -- local dd = DATA:Get();
        -- print("data: ", #dd);
        REQ.weakFilter.myWishList = LFG:GetCheckedDungeonsHash();
        f.faux:UpdateButtons();
    end);

    f:SetScript("OnHide", function(self) 
        --DATA.OnAdd = nil;
    end);

    DATA:AddOnPlayerUpdateHandler(function(player) 
        if not f:IsShown() then return end        
        f.faux:UpdateButtons();
        --f.faux:UpdateButtonByItem(player);
    end)

    REQ:AddOnFilterChange(function()
        if not f:IsShown() then return end
        f.faux:UpdateButtons();
    end);

    return f;
end

function A:GetLfmReq()
    return REQ;
end