local A = select(2, ...);
local fb = A.Lib.FB;
local C = A.Constants;
local Cache = A.Cache;
local U = A.Utils;
local L = A.Localize.translate;

local TabButtonMixin = {
    OnClick = function(self)
        PanelTemplates_Tab_OnClick(self, self:GetParent());
        self:GetParent():ShowTab(self:GetID());
        PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);
    end
}

-- coommon builders for tab frame
local TabBuilders = {

    -- positioning
    function(f)
        local parent = f:GetParent();
        f:SetPoint("TOPLEFT", parent, "TOPLEFT");
        f:SetPoint("BOTTOMRIGHT", parent, "BOTTOMRIGHT");
    end,

    -- tab frame header
    function(f)
        local t = f:CreateFontString("$parent_Header", "OVERLAY", "GameFontNormal");
        t:SetPoint("TOPLEFT", f, "TOPLEFT", 70, 3);
        t:SetText(C.Frames[f.frameName].Header);
        t:SetSize(300, 30);
        t:SetJustifyH("LEFT");
    end,

    -- send request button
    -- function(f)

    --     local lfgb = CreateFrame("Button", "$parent_SendRequest", f, "UIPanelButtonTemplate");
    --     lfgb:SetText(C.SendRequest);
    --     lfgb:SetScript("OnClick", function() 
    --         SendChatMessage(P:ToRequestMessage() ,"Say");
    --     end);
    --     lfgb:SetPoint("BOTTOM", f, "BOTTOM", 0, 4);
    --     lfgb:SetWidth(200);
    
    -- end
}


local MainFrameMixin = {
    CreateTabButton = function(self, id, text, point)
        local t = CreateFrame("Button", "$parentTab" .. id, self, "CharacterFrameTabButtonTemplate");
        t:SetID(id);
        t:SetText(text);
        t:SetPoint(unpack(point));

        Mixin(t, TabButtonMixin);
        t:SetScript("OnClick", t.OnClick);
        return t;
    end,

    CreateTab = function(self, id, builder)
        if not self.tabFrames then
            self.tabFrames = {};
        end

        local tab = builder(self, id);
        fb:ApplyFrameBuilders(tab, TabBuilders);
        table.insert(self.tabFrames, tab);

    end,

    ShowTab = function(self, id)
        for _, tab in pairs(self.tabFrames) do
            if tab.tabId == id then
                tab:Show();
            else
                tab:Hide();
            end
        end
        PanelTemplates_SetTab(self, id);
    end
}

local Builders = {

    -- main things
    function(f)
        f:SetSize(600, 500);
        f:SetPoint("CENTER");
        fb:MakeMovable(f);
        local texture = _G[f:GetName().."Portrait"];
        texture:SetTexture("Interface\\LFGFRAME\\UI-LFG-PORTRAIT"); --(LFG_EYE_TEXTURES.default); --("")
        --SetPortraitTexture(texture, "player");
    end,

    -- creating lfm and lfg tabs
    function(f)
        f:CreateTab(1, function(p, id) return A.Lfg:BuildFrame(p, id) end);
        f:CreateTab(2, function(p, id) return A.Lfm:BuildFrame(p, id) end);
    end,

    -- create tab buttons
    function(f)
        local prev = f:CreateTabButton(1, C.Lfg, {"TOPLEFT", f, "BOTTOMLEFT", 10, 0})
        f:CreateTabButton(2, C.Lfm, {"LEFT", prev, "RIGHT", 0, 0});
        PanelTemplates_SetNumTabs(f, 2);
        f:ShowTab(1);
    end,

    function(f)
        local t = f:CreateFontString("$parent_Version", "OVERLAY", "GameFontNormalSmall");
        t:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", -12, 0);
        t:SetText(U:InlineColor("v" .. A.BUILD.appVersion, "ffffd0"));
        t:SetSize(60, 30);
        t:SetJustifyH("RIGHT");
    end,

    -- Companion options
    function(b)
        local check = CreateFrame("CheckButton", "$parent_ConpanionToggleButton", b, "ChatConfigCheckButtonTemplate");
        
        --b.checked = check;
        check:SetPoint("BOTTOMLEFT", b, "BOTTOMLEFT", 8, 3);
        check:SetHitRectInsets(0, -200, 0, 0);
        check:SetChecked(A.Companion:IsShown());
        check:SetScript("OnClick", function(self, ...) 
            if self:GetChecked() then
                A.Companion:Show();
            else
                A.Companion:Hide();
            end
        end);
        local t = check:CreateFontString("$parent_Text", "OVERLAY", "GameFontNormal");
        t:SetPoint("LEFT", check, "RIGHT");
        t:SetText(L("Show Tabu-Lfg Companion"));

        --check:SetScript("OnClick", OnButtonChecked);

        return t;
    end,
    function(b, prev)
        local check = CreateFrame("CheckButton", "$parent_ConpanionAutoshowButton", b, "ChatConfigCheckButtonTemplate");
        
        --b.checked = check;
        check:SetPoint("LEFT", prev, "RIGHT", 20, 0);
        check:SetHitRectInsets(0, -200, 0, 0);
        check:SetChecked(Cache:GetOption("companion.show") == true);
        check:SetScript("OnClick", function(self, ...) 
            Cache:SetOption("companion.show", self:GetChecked());
        end);
        local t = check:CreateFontString("$parent_Text", "OVERLAY", "GameFontNormal");
        t:SetPoint("LEFT", check, "RIGHT");
        t:SetText(L("Autoshow Tabu-Lfg Companion on login"));

        --check:SetScript("OnClick", OnButtonChecked);

        return check;
    end,
}

function A:CreateMainFrame()
    local f = CreateFrame("Frame", "TabuLfg", UIParent, "ButtonFrameTemplate"); --"UIPanelDialogTemplate"); -- "UIPanelDialogTemplate");
    Mixin(f, MainFrameMixin);
    fb:ApplyFrameBuilders(f, Builders);
    f:Hide();

    return f;
end

function A:GetMainFrame()
    local f = A.mainFrame;
    if not f then
        f = self:CreateMainFrame();
        A.mainFrame = f;
    end
    return f;
end

function A:ShowMainFrame()
    local f = self:GetMainFrame();
    f:Show();
end

function A:HideMainFrame()
    local f = self:GetMainFrame();
    f:Hide();
end

function A:ToggleMainFrame()
    local f = self:GetMainFrame();
    if f:IsShown() then
        f:Hide();
    else
        f:Show();
    end
end